from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from typing import Type

import isodate
import jsonschema
import pandas as pd
import pydantic

from bagel import dataset_description_model, dictionary_models, mappings
from bagel.logger import log_error, logger
from bagel.mappings import DEPRECATED_NAMESPACE_PREFIXES, NB

# TODO: Once we remove support for v1 annotation tool data dictionaries, revert to using this version for data dictionary schema validation
# DICTIONARY_SCHEMA = dictionary_models.DataDictionary.model_json_schema()

AGE_FORMATS = {
    "float": NB.pf + ":FromFloat",
    "int": NB.pf + ":FromInt",
    "euro": NB.pf + ":FromEuro",
    "bounded": NB.pf + ":FromBounded",
    "iso8601": NB.pf + ":FromISO8601",
    "range": NB.pf + ":FromRange",
}


def get_available_configs(config_namespaces_mapping: list) -> list:
    """Return the list of names of available community configurations."""
    return [config["config_name"] for config in config_namespaces_mapping]


def additional_config_help_text() -> str:
    """
    Construct a warning to be added in the help text for the --config option when no configurations are available.
    This is to inform the user explicitly in the rare case when no configurations are available, i.e. when fetching the remote file fails and there is no local backup found
    (if submodules were not properly initialized). In this case, allowed choices for --config would be shown as an empty list [],
    meaning even the default "Neurobagel" option would result in an invalid parameter error (and the `pheno` command cannot run).
    """
    if mappings.CONFIG_NAMESPACES_MAPPING == []:
        return (
            "[bold red]WARNING: Failed to locate any community configurations. "
            "Please check that you have an internet connection, or open an issue in https://github.com/neurobagel/bagel-cli/issues if the problem persists.[/bold red]"
        )
    return ""


def check_if_remote_config_namespaces_used():
    """Warn if the community configuration namespaces could not be fetched from the remote source."""
    if (
        mappings.CONFIG_NAMESPACES_FETCHING_ERR
        and mappings.CONFIG_NAMESPACES_MAPPING != []
    ):
        logger.warning(
            f"Failed to fetch configuration from {mappings.CONFIG_NAMESPACES_URL}. Error: {mappings.CONFIG_NAMESPACES_FETCHING_ERR}. "
            "Using a packaged backup configuration instead *which may be outdated*. "
            "Check your internet connection?"
        )


def get_supported_namespaces_for_config(config_name: str) -> dict:
    """Return a dictionary of supported namespace prefixes and their corresponding full URLs for a given community configuration."""
    config_namespaces = next(
        config["namespaces"]
        for config in mappings.CONFIG_NAMESPACES_MAPPING
        if config["config_name"] == config_name
    )

    config_namespaces_dict = {}
    for namespace_group in config_namespaces.values():
        for namespace in namespace_group:
            config_namespaces_dict[namespace["namespace_prefix"]] = namespace[
                "namespace_url"
            ]

    return config_namespaces_dict


def get_columns_about(data_dict: dict, concept: str) -> list:
    """
    Returns all column names that have been annotated as "IsAbout" the desired concept.
    Parameters
    ----------
    data_dict: dict
        A valid Neurobagel annotated data dictionary must be provided.
    concept: str
        A (shorthand) IRI for a concept that a column can be "about"

    Returns
    list
        List of column names that are "about" the desired concept

    -------

    """
    return [
        col
        for col, content in get_annotated_columns(data_dict)
        if content["Annotations"]["IsAbout"]["TermURL"] == concept
    ]


def get_annotated_columns(data_dict: dict) -> list[tuple[str, dict]]:
    """
    Return a list of all columns that have Neurobagel 'Annotations' in a data dictionary,
    where each annotated column is represented as a tuple of the column name (dictionary key from the data dictionary) and
    properties (all dictionary contents from the data dictionary).
    """
    return [
        (col, content)
        for col, content in data_dict.items()
        if "Annotations" in content
    ]


def recursive_find_values_for_key(data: dict, target: str) -> list:
    """
    Recursively search for a key in a possibly nested dictionary and return a list of all values found for that key.

    TODO: This function currently only considers nested dicts, and would need to be expanded if Neurobagel
    data dictionaries grow to have controlled terms inside list objects.
    """
    target_values = []
    if isinstance(data, dict):
        for key, value in data.items():
            if key == target:
                target_values.append(value)
            else:
                target_values.extend(
                    recursive_find_values_for_key(data=value, target=target)
                )
    return target_values


def find_unsupported_namespaces_and_term_urls(
    data_dict: dict,
    config: str,
) -> tuple[list, dict]:
    """
    From a provided data dictionary, find all term URLs that contain an unsupported namespace prefix.
    Return a tuple of unsupported prefixes and a dictionary of the offending column names and their unrecognized term URLs.
    """
    unsupported_prefixes = set()
    unrecognized_term_urls = {}

    supported_namespaces = get_supported_namespaces_for_config(config)

    for col, content in get_annotated_columns(data_dict):
        for col_term_url in recursive_find_values_for_key(
            content["Annotations"], "TermURL"
        ):
            prefix = col_term_url.split(":")[0]
            if prefix not in supported_namespaces:
                unsupported_prefixes.add(prefix)
                unrecognized_term_urls[col] = col_term_url

    # sort the prefixes for a predictable order in the error message
    return sorted(unsupported_prefixes), unrecognized_term_urls


def find_deprecated_namespaces(namespaces: list) -> list:
    """Return the deprecated vocabulary namespace prefixes found in a list of namespace prefixes."""
    return [ns for ns in namespaces if ns in DEPRECATED_NAMESPACE_PREFIXES]


def map_std_vars_to_columns(data_dict: dict) -> dict[str, list]:
    """
    Maps all Neurobagel standardized variables (e.g. "Sex") to a list containing all column names (if any) that
    have been linked to this variable.

    Returns a dictionary where the keys are the aliases for Neurobagel standardized variables and the values are lists of column names.
    """
    return {
        std_var_name: get_columns_about(data_dict, std_var_iri)
        for std_var_name, std_var_iri in mappings.NEUROBAGEL.items()
        if get_columns_about(data_dict, std_var_iri)
    }


def map_tools_to_columns(data_dict: dict) -> dict[str, list]:
    """
    Return a mapping of all assessment tools described in the data dictionary to the columns that
    are mapped to it.

    Returns a dictionary where the keys are the assessment tool IRIs and the values are lists of column names.
    """
    out_dict = defaultdict(list)
    for col, content in get_annotated_columns(data_dict):
        part_of = content["Annotations"].get("IsPartOf")
        if part_of is not None:
            out_dict[part_of.get("TermURL")].append(col)

    return out_dict


def is_missing_value(value: str | int, column: str, data_dict: dict) -> bool:
    """Determine if a raw value is listed as a missing value in the data dictionary entry for this column"""
    return value in data_dict[column]["Annotations"].get("MissingValues", [])


def is_column_type(
    column: str,
    data_dict: dict,
    variable_type: Type[dictionary_models.Neurobagel],
) -> bool:
    """Determine whether a column annotation in a Neurobagel data dictionary fits the specified variable type."""
    column_annotation = data_dict[column]["Annotations"]

    try:
        variable_type.model_validate(column_annotation)
        return True
    except pydantic.ValidationError:
        return False


def map_cat_val_to_term(value: str | int, column: str, data_dict: dict) -> str:
    """Take a raw categorical value and return the controlled term it has been mapped to"""
    return data_dict[column]["Annotations"]["Levels"][value]["TermURL"]


def get_age_format(column: str, data_dict: dict) -> str:
    return data_dict[column]["Annotations"]["Format"]["TermURL"]


def transform_age(value: str, value_format: str) -> float:
    try:
        if value_format in [
            AGE_FORMATS["float"],
            AGE_FORMATS["int"],
        ]:
            return float(value)
        if value_format == AGE_FORMATS["euro"]:
            return float(value.replace(",", "."))
        if value_format == AGE_FORMATS["bounded"]:
            return float(value.strip("+"))
        if value_format == AGE_FORMATS["iso8601"]:
            if not value.startswith("P"):
                pvalue = "P" + value
            else:
                pvalue = value
            duration = isodate.parse_duration(pvalue)
            return float(duration.years + duration.months / 12)
        if value_format == AGE_FORMATS["range"]:
            a_min, a_max = value.split("-")
            return sum(map(float, [a_min, a_max])) / 2
        log_error(
            logger,
            f"The data dictionary contains an unrecognized age format: {value_format}. "
            f"Ensure that the format TermURL is one of {list(AGE_FORMATS.values())}.",
        )
    except (ValueError, isodate.isoerror.ISO8601Error) as e:
        log_error(
            logger,
            f"Error applying the format {value_format} to the age value: {value}. Error: {e}\n"
            f"Check your data dictionary to ensure that the annotated age format matches the age values in your phenotypic table, "
            "and that any missing values in your age column have been correctly annotated. "
            "For examples of acceptable values for specific age formats, see https://neurobagel.org/data_models/dictionaries/#age.",
        )


def get_transformed_values(
    columns: list,
    row: pd.Series,
    data_dict: dict,
) -> list:
    """
    Convert a list of raw phenotypic values to the corresponding controlled terms.
    NOTE: Assumes that the values come from columns that have not been annotated as being about an assessment tool.
    """
    transf_vals: list[float | str] = []
    for col in columns:
        value = row[col]
        if is_missing_value(value, col, data_dict):
            continue
        if is_column_type(
            col, data_dict, dictionary_models.CategoricalNeurobagel
        ):
            transf_vals.append(map_cat_val_to_term(value, col, data_dict))
        else:
            # TODO: replace with more flexible solution when we have more
            # continuous variables than just age
            transf_vals.append(
                transform_age(str(value), get_age_format(col, data_dict))
            )

    return transf_vals


def get_transformed_row_for_table(
    columns: list,
    row: pd.Series,
    data_dict: dict,
    collection_mapping: dict,
) -> dict:
    """
    For specified columns within a row of a phenotypic table, create a dict where keys are the standardized variable IDs
    corresponding to columns, and values are standardized representations of the raw column values.
    - Continuous and categorical columns are transformed to their standardized values.
    - For columns belonging to a collection, a single standardized column is created for the collection
      indicating whether data is available for at least one column in the collection.
    - Identifier column values are left unchanged.
    """
    collection_available_term = "nb:available"
    collection_unavailable_term = "nb:unavailable"

    transformed_row: dict[str, str | float] = {}
    for col in columns:
        if is_column_type(
            col, data_dict, dictionary_models.CollectionNeurobagel
        ):
            continue
        std_var_name = data_dict[col]["Annotations"]["IsAbout"]["TermURL"]
        value = row[col]
        if is_missing_value(value, col, data_dict):
            transformed_row[std_var_name] = ""
        elif is_column_type(
            col, data_dict, dictionary_models.IdentifierNeurobagel
        ):
            transformed_row[std_var_name] = str(value)
        elif is_column_type(
            col, data_dict, dictionary_models.CategoricalNeurobagel
        ):
            transformed_row[std_var_name] = map_cat_val_to_term(
                value, col, data_dict
            )
        else:
            # TODO: replace with more flexible solution when we have more
            # continuous variables than just age
            transformed_row[std_var_name] = transform_age(
                str(value), get_age_format(col, data_dict)
            )

    if collection_mapping:
        for collection, item_cols in collection_mapping.items():
            if are_any_available(item_cols, row, data_dict):
                transformed_row[collection] = collection_available_term
            else:
                transformed_row[collection] = collection_unavailable_term

    return transformed_row


# TODO: Check all columns and then return list of offending columns' names
def categorical_cols_have_bids_levels(data_dict: dict) -> bool:
    for col, content in get_annotated_columns(data_dict):
        if (
            is_column_type(
                col, data_dict, dictionary_models.CategoricalNeurobagel
            )
            and content.get("Levels") is None
        ):
            return False

    return True


def get_mismatched_categorical_levels(data_dict: dict) -> list:
    """
    Returns list of any categorical columns from a data dictionary that have different entries
    for the "Levels" key between the column's BIDS and Neurobagel annotations.
    """
    mismatched_cols = []
    for col, content in get_annotated_columns(data_dict):
        if is_column_type(
            col, data_dict, dictionary_models.CategoricalNeurobagel
        ):
            known_levels = list(
                content["Annotations"]["Levels"].keys()
            ) + content["Annotations"].get("MissingValues", [])
            if set(content.get("Levels", {}).keys()).difference(known_levels):
                mismatched_cols.append(col)

    return mismatched_cols


def are_any_available(columns: list, row: pd.Series, data_dict: dict) -> bool:
    """
    Checks that at least one of the values in the specified columns is not a missing value.
    This is mainly useful to determine the availability of an assessment tool
    """
    return any(
        not is_missing_value(value, column, data_dict)
        for column, value in row[columns].items()
    )


def find_missing_annotated_cols(
    data_dict: dict, pheno_df: pd.DataFrame
) -> list:
    """
    Find columns that are annotated in the data dictionary but not present in the phenotypic file.
    """
    missing_annotated_cols = []
    for col, _ in get_annotated_columns(data_dict):
        if col not in pheno_df.columns:
            missing_annotated_cols.append(col)

    return missing_annotated_cols


def find_undefined_cat_col_values(
    data_dict: dict, pheno_df: pd.DataFrame
) -> dict[str, list]:
    """
    Checks that all categorical column values have annotations. Returns a dictionary containing
    any categorical column names and specific column values not defined in the corresponding data
    dictionary entry.
    """
    all_undefined_values = {}
    for col, content in get_annotated_columns(data_dict):
        if is_column_type(
            col, data_dict, dictionary_models.CategoricalNeurobagel
        ):
            known_values = list(
                content["Annotations"]["Levels"].keys()
            ) + content["Annotations"].get("MissingValues", [])
            unknown_values = []
            for value in pheno_df[col].unique():
                if value not in known_values:
                    unknown_values.append(value)
            if unknown_values:
                all_undefined_values[col] = unknown_values

    return all_undefined_values


def find_unused_missing_values(
    data_dict: dict, pheno_df: pd.DataFrame
) -> dict[str, list]:
    """
    Checks if missing values annotated in the data dictionary appear at least once in the phenotypic file.
    Returns a dictionary containing any column names and annotated missing values not found in the phenotypic
    file column.
    """
    all_unused_missing_vals = {}
    for col, content in get_annotated_columns(data_dict):
        unused_missing_vals = []
        for missing_val in content["Annotations"].get("MissingValues", []):
            if missing_val not in pheno_df[col].unique():
                unused_missing_vals.append(missing_val)
        if unused_missing_vals:
            all_unused_missing_vals[col] = unused_missing_vals

    return all_unused_missing_vals


def get_rows_with_empty_strings(df: pd.DataFrame, columns: list) -> list:
    """For specified columns, returns the indices of rows with empty strings"""
    # NOTE: Profile this section if things get slow, transforming "" -> nan and then
    # using .isna() will very likely be much faster
    empty_row = df[columns].eq("").any(axis=1)
    # Return the row index as it would look in a spreadsheet program, 1-based and including the header
    return [idx + 2 for idx in empty_row[empty_row].index]


def construct_dictionary_schema_for_validation() -> dict:
    """
    For backwards compatibility with the v1 annotation tool, we patch the data dictionary schema
    to allow for either a 'Format' or 'Transformation' key (but not both) for continuous columns.
    This ensures that v1 annotation tool data dictionaries still pass the initial schema validation (for now),
    without encoding the 'Transformation' key in the updated data dictionary model itself.

    TODO: Remove once we no longer support data dictionaries annotated using annotation tool v1.
    """
    patched_schema = deepcopy(
        dictionary_models.DataDictionary.model_json_schema()
    )
    continuous_schema = patched_schema["$defs"]["ContinuousNeurobagel"]
    continuous_schema["properties"]["Transformation"] = deepcopy(
        continuous_schema["properties"]["Format"]
    )
    if "Format" in continuous_schema.get("required", []):
        continuous_schema["required"].remove("Format")
    continuous_schema["oneOf"] = [
        {"required": ["Transformation"], "not": {"required": ["Format"]}},
        {"required": ["Format"], "not": {"required": ["Transformation"]}},
    ]
    return patched_schema


def validate_data_dict(data_dict: dict, config: str | None) -> None:
    try:
        jsonschema.validate(
            data_dict, construct_dictionary_schema_for_validation()
        )
    except jsonschema.ValidationError as e:
        # TODO: When *every* item in an input JSON is not schema valid,
        # jsonschema.validate will raise a ValidationError for only *one* item among them.
        # Weirdly, the item that is chosen can vary, possibly due to jsonschema internally processing/validating
        # items in an inconsistent order.
        # You can reproduce this by running `bagel pheno` on the example_invalid input pair.
        # This isn't necessarily a problem as the error will not be wrong, but if we care about
        # returning ALL invalid items, we may want to use something like a Draft7Validator instance instead.
        # NOTE: If the validation error occurs at the root level (i.e., the entire JSON object fails),
        # e.path may be empty. We have a backup descriptor "Entire document" for the offending item in this case.
        log_error(
            logger,
            "The data dictionary is not a valid Neurobagel data dictionary. "
            f"Entry that failed validation: {e.path[-1] if e.path else 'Entire document'}\n"
            f"Details: {e.message}\n"
            "[italic]TIP: Ensure each annotated column contains an 'Annotations' key.[/italic]",
        )

    if get_annotated_columns(data_dict) == []:
        log_error(
            logger,
            "The data dictionary must contain at least one column with Neurobagel annotations.",
        )

    if config is not None:
        unsupported_namespaces, unrecognized_term_urls = (
            find_unsupported_namespaces_and_term_urls(data_dict, config)
        )
        if unsupported_namespaces:
            namespace_deprecation_msg = ""
            if config == mappings.DEFAULT_CONFIG:
                if deprecated_namespaces := find_deprecated_namespaces(
                    unsupported_namespaces
                ):
                    namespace_deprecation_msg = (
                        f"\n\nMore info: The following vocabularies have been deprecated by Neurobagel: {deprecated_namespaces}. "
                        "Please update your data dictionary using the latest version of the annotation tool at https://annotate.neurobagel.org."
                    )
            log_error(
                logger,
                f"The data dictionary contains unsupported vocabulary namespace prefixes: {unsupported_namespaces}\n"
                f"Unsupported vocabularies are used for terms in the following columns' annotations: {unrecognized_term_urls}\n"
                f"Please ensure the data dictionary only includes terms from vocabularies recognized by {config}. "
                f"{namespace_deprecation_msg}",
            )

    if (
        len(
            get_columns_about(
                data_dict, concept=mappings.NEUROBAGEL["participant"]
            )
        )
        == 0
    ):
        log_error(
            logger,
            "The data dictionary must contain at least one column annotated as being about participant ID.",
        )

    # TODO: remove this validation when we start handling multiple participant and / or session ID columns
    if (
        len(
            get_columns_about(
                data_dict, concept=mappings.NEUROBAGEL["participant"]
            )
        )
        > 1
    ) | (
        len(
            get_columns_about(
                data_dict, concept=mappings.NEUROBAGEL["session"]
            )
        )
        > 1
    ):
        log_error(
            logger,
            "The data dictionary has more than one column about participant ID or session ID. "
            "Please ensure only one column is annotated for participant and session IDs.",
        )

    if (
        set(map_std_vars_to_columns(data_dict).keys())
        == {"participant", "session"}
    ) or (set(map_std_vars_to_columns(data_dict).keys()) == {"participant"}):
        logger.warning(
            "The only columns annotated in the data dictionary are participant ID or session ID columns. "
            "As a result, the generated graph-ready data will not contain any subject phenotypic characteristics. "
            "Check that all relevant phenotypic columns in your data table have been annotated."
        )

    if (
        len(get_columns_about(data_dict, concept=mappings.NEUROBAGEL["sex"]))
        > 1
    ):
        logger.warning(
            "The data dictionary indicates more than one column about sex. "
            "Neurobagel cannot resolve multiple sex values per subject-session, and so will only consider the first identified column for sex data."
        )

    if (
        len(get_columns_about(data_dict, concept=mappings.NEUROBAGEL["age"]))
        > 1
    ):
        logger.warning(
            "The data dictionary indicates more than one column about age. "
            "Neurobagel cannot resolve multiple age values per subject-session, and so will use only the first identified column for age data."
        )

    if (
        len(
            get_columns_about(
                data_dict, concept=mappings.NEUROBAGEL["subject_group"]
            )
        )
        > 1
    ):
        logger.warning(
            "The data dictionary indicates more than one column about subject group. "
            "Neurobagel cannot resolve multiple subject group values per subject-session, and so will use only the first identified column for subject group data."
        )

    if not categorical_cols_have_bids_levels(data_dict):
        logger.warning(
            "The data dictionary contains at least one column that looks categorical but lacks a BIDS 'Levels' attribute."
        )

    if mismatched_cols := get_mismatched_categorical_levels(data_dict):
        logger.warning(
            f"The data dictionary contains columns with mismatched levels between the BIDS and Neurobagel annotations: {mismatched_cols}"
        )


def check_for_duplicate_ids(data_dict: dict, pheno_df: pd.DataFrame):
    """Log an error if there are duplicate participant IDs or duplicate combinations of participant and session IDs, if both are present."""
    id_columns = get_columns_about(
        data_dict, concept=mappings.NEUROBAGEL["participant"]
    ) + get_columns_about(data_dict, concept=mappings.NEUROBAGEL["session"])
    duplicates_mask = pheno_df.duplicated(subset=id_columns, keep=False)
    if duplicates_mask.any():
        duplicate_indices = [
            idx + 2 for idx in pheno_df[duplicates_mask].index
        ]
        log_error(
            logger,
            "The phenotypic table contains duplicate participant IDs or duplicate combinations of participant and session IDs. "
            f"Duplicate IDs were found in these rows (header row is 1): {duplicate_indices}. "
            "Ensure that each row represents a unique participant or participant-session (if a session column is present).",
        )


def get_validated_dataset_description_and_incomplete_fields(
    raw_dataset_desc: dict,
) -> tuple[dataset_description_model.DatasetDescription, set]:
    """
    Return a validated dataset description instance and any incomplete optional fields (by their aliases),
    including those were unset or those set to empty or whitespace-only strings.
    """
    validated_dataset_desc = (
        dataset_description_model.DatasetDescription.model_validate(
            raw_dataset_desc
        )
    )
    model_fields = [
        field.alias
        for field in dataset_description_model.DatasetDescription.model_fields.values()
    ]
    non_default_fields = validated_dataset_desc.model_dump(
        by_alias=True, exclude_defaults=True
    ).keys()
    incomplete_fields = set(model_fields) - set(non_default_fields)
    return validated_dataset_desc, incomplete_fields


def validate_dataset_description(
    dataset_desc: dict,
) -> dataset_description_model.DatasetDescription:
    """
    Return a validated instance of the dataset description and log an informative warning if any optional fields are missing.
    """
    try:
        validated_dataset_desc, unset_fields = (
            get_validated_dataset_description_and_incomplete_fields(
                dataset_desc
            )
        )
        if unset_fields:
            logger.warning(
                "The dataset description is missing or has empty values for the following recommended fields:\n"
                + "\n".join(f"- {field}" for field in unset_fields)
                + "\nConsider completing these fields to improve dataset reuse and clarify access procedures."
                + "\nSee the documentation at https://neurobagel.org/user_guide/dataset_description/"
            )
        return validated_dataset_desc
    except pydantic.ValidationError as err:
        # Pydantic will list all offending fields and their issues in the error message (https://docs.pydantic.dev/latest/errors/errors/)
        log_error(
            logger,
            "The dataset description is invalid. "
            "\nSee the documentation at https://neurobagel.org/user_guide/dataset_description/. "
            f"\nValidation details:\n"
            f"{err}",
        )


def validate_inputs(
    data_dict: dict, pheno_df: pd.DataFrame, config: str | None = None
) -> None:
    """Determine whether the input data dictionary and phenotypic table are valid and compatible."""
    validate_data_dict(data_dict, config)

    if missing_annotated_cols := find_missing_annotated_cols(
        data_dict, pheno_df
    ):
        log_error(
            logger,
            "The provided phenotypic table and data dictionary are incompatible. "
            f"The following columns are annotated in the data dictionary but are missing from the phenotypic table: {missing_annotated_cols}. "
            "Check that you've selected the correct data dictionary for your phenotypic table. "
            "Each column described in the data dictionary must have a corresponding column with the same name in the phenotypic table.",
        )

    columns_about_ids = get_columns_about(
        data_dict, concept=mappings.NEUROBAGEL["participant"]
    ) + get_columns_about(data_dict, concept=mappings.NEUROBAGEL["session"])
    if row_indices := get_rows_with_empty_strings(pheno_df, columns_about_ids):
        log_error(
            logger,
            "The phenotypic table contains missing values in participant or session ID columns. "
            "Ensure that each row includes a non-empty participant ID (and session ID, if the table contains a session ID column). "
            f"Missing IDs were found in these rows (header row is 1): {row_indices}. "
            "[italic]TIP: Check that your table does not have any completely empty rows.[/italic]",
        )

    check_for_duplicate_ids(data_dict, pheno_df)

    undefined_cat_col_values = find_undefined_cat_col_values(
        data_dict, pheno_df
    )
    if undefined_cat_col_values:
        log_error(
            logger,
            "One or more unique values found in annotated categorical columns of the phenotypic table are missing annotations in the data dictionary "
            rf"(shown by column as 'column_name': \[unannotated_values]): {undefined_cat_col_values}. "
            "Check that you've selected the correct data dictionary or annotate the values that are missing. "
            "[italic]TIP: Ensure that column values in the table exactly match the values annotated in the data dictionary.[/italic]",
        )

    unused_missing_values = find_unused_missing_values(data_dict, pheno_df)
    if unused_missing_values:
        logger.warning(
            "Some values annotated as missing values in the data dictionary were not found "
            r"in the corresponding phenotypic table column(s) (shown as 'column_name': \[unused_missing_values]): "
            f"{unused_missing_values}. [italic]If this is not intentional, please check your data dictionary "
            "and phenotypic table.[/italic]"
        )


def convert_transformation_to_format(data_dict: dict) -> dict:
    """
    If the uploaded data dictionary contains any "Transformation" keys, rename the key(s) to "Format"
    internally for downstream operations involving the data dictionary.
    This ensures compatibility with both v1 and v2 annotation tool-generated data dictionaries.

    TODO: Remove when we no longer support data dictionaries annotated using annotation tool v1.
    """
    age_column_names = get_columns_about(
        data_dict, concept=mappings.NEUROBAGEL["age"]
    )
    if age_column_names:
        age_cols_using_transformation = []
        for age_column_name in age_column_names:
            age_annotations = data_dict[age_column_name]["Annotations"]
            if "Transformation" in age_annotations:
                age_cols_using_transformation.append(age_column_name)
                age_annotations["Format"] = age_annotations.pop(
                    "Transformation"
                )

        if age_cols_using_transformation:
            logger.warning(
                f"The data dictionary contains a deprecated 'Transformation' key in the annotations for the column(s): {age_cols_using_transformation}. "
                "This key has been replaced by 'Format'. For now, 'Transformation' will be interpreted as equivalent to 'Format', "
                "but support for 'Transformation' will be removed in a future release. "
                "We recommend updating your data dictionary using the latest version of the Neurobagel annotation tool."
            )

    return data_dict


def dataset_description_to_graph_attributes(
    dataset_description: dataset_description_model.DatasetDescription,
) -> dict:
    """Convert fields from a dataset description to JSONLD-ready attributes for a dataset."""
    dataset_graph_attributes = {
        "hasLabel": dataset_description.name,
        "hasAuthors": (
            dataset_description.authors
            if dataset_description.authors
            else None
        ),
        "hasReferencesAndLinks": (
            dataset_description.references_and_links
            if dataset_description.references_and_links
            else None
        ),
        "hasKeywords": (
            dataset_description.keywords
            if dataset_description.keywords
            else None
        ),
        "hasRepositoryURL": dataset_description.repository_url,
        "hasAccessInstructions": dataset_description.access_instructions,
        "hasAccessType": dataset_description.access_type,
        "hasAccessEmail": dataset_description.access_email,
        "hasAccessLink": dataset_description.access_link,
    }

    return dataset_graph_attributes



# This is an example script for generating a JSON schema for the Neurobagel graph data model (i.e., the model for a Neurobagel "Dataset").
# Example usage: python generate_graphdata_jsonschema.py

import json

from bagel.models import Dataset

FPATH = "neurobagel_graphdata.schema.json"

with open(FPATH, "w") as f:
    f.write(
        json.dumps(
            Dataset.model_json_schema(),
            indent=2,
        )
    )



import json
from pathlib import Path
from typing import Any

import httpx
import pandas as pd
import typer

from ..logger import log_error, logger


def file_encoding_error_message(input_p: Path) -> str:
    """Return a message for when a file cannot be read due to encoding issues."""
    return (
        f"Failed to decode the input file {input_p}. "
        "Please ensure that all input files have UTF-8 encoding. "
        "[italic]TIP: Need help converting your file? Try a tool like iconv (http://linux.die.net/man/1/iconv) or https://www.freeformatter.com/convert-file-encoding.html.[/italic]"
    )


# TODO: Consider adding a function parameter to allow enabling `keep_default_na`.
# For some tables the CLI reads (e.g., BIDS tables), parsing empty strings as NaN is useful
# and saves us from needing additional custom string validation to catch empty cells.
# See https://github.com/neurobagel/bagel-cli/issues/505
def load_tabular(
    input_p: Path, input_type: str = "phenotypic"
) -> pd.DataFrame:
    """Load a .tsv pheno file and do some basic validation of the file type."""
    if input_p.suffix == ".tsv":
        try:
            tabular_df = pd.read_csv(
                input_p,
                sep="\t",
                keep_default_na=False,
                dtype=str,
                encoding="utf-8",
            )
        except UnicodeDecodeError:
            log_error(
                logger,
                file_encoding_error_message(input_p),
            )

        if tabular_df.shape[1] > 1:
            return tabular_df

        # If we have only one column, but splitting by ',' gives us several elements
        # then there is a good chance the user accidentally renamed a .csv into .tsv
        # and we should give them some extra info with our error message to fix this.
        note_misnamed_csv = (
            f"[italic]NOTE: Your {input_type} table resembles a .csv file "
            "as it contains several ',' commas. Check that you have not "
            "accidentally renamed a .csv file to a .tsv.[/italic]"
        )
        log_error(
            logger,
            f"{input_p} is not a valid Neurobagel {input_type} table (.tsv). "
            "This file is expected to have multiple columns but only one column was found. "
            f"\n{note_misnamed_csv if len(tabular_df.columns[0].split(',')) > 1 else ''}",
        )

    log_error(
        logger,
        f"({input_p}) is not a .tsv file. "
        f"Please provide a valid .tsv {input_type} table!",
    )


def load_json(input_p: Path) -> Any:
    """Load a user-specified json type file."""
    try:
        with open(input_p, "r", encoding="utf-8") as f:
            return json.load(f)
    except UnicodeDecodeError:
        log_error(
            logger,
            file_encoding_error_message(input_p),
        )
    except json.JSONDecodeError:
        log_error(
            logger,
            f"File is not valid JSON: {input_p}. Please provide a valid JSON file.",
        )


def check_overwrite(output: Path, overwrite: bool):
    """Exit program gracefully if an output file already exists but --overwrite has not been set."""
    if output.exists() and not overwrite:
        raise typer.Exit(
            typer.style(
                f"Output file {output} already exists. Use --overwrite or -f to overwrite.",
                fg=typer.colors.RED,
            )
        )


def save_jsonld(data: dict, filename: Path):
    with open(filename, "w", encoding="utf-8") as f:
        f.write(json.dumps(data, indent=2, ensure_ascii=False))
    logger.info(f"Saved output to:  {filename}")


def request_file(url: str, backup_path: Path) -> tuple[list, str | None]:
    contents = []
    err = None

    try:
        response = httpx.get(url)
        response.raise_for_status()
        contents = response.json()
    except (httpx.HTTPError, json.JSONDecodeError) as request_err:
        try:
            # We don't use file_utils.load_json() here because we don't want to throw an error yet if there are problems with the file.
            # Otherwise, since this function is called to populate global constants in mappings.py, there may be exceptions on import,
            # meaning there will be errors even if the user runs just bagel --help.
            with open(backup_path, "r", encoding="utf-8") as f:
                contents = json.load(f)
                err = str(request_err)
        except Exception as load_err:
            err = str(load_err)

    return contents, err



from typing import Iterable

import pandas as pd

from bagel import mappings, models
from bagel.logger import log_error, logger

# Shorthands for expected column names in a Nipoppy processing status file
# TODO: While there are multiple session ID columns in a Nipoppy processing status file,
# we only only look at `bids_session_id` right now. We should revisit this after the schema is finalized,
# to see if any other logic is needed to avoid issues with session ID discrepancies across columns.
PROC_STATUS_COLS = {
    "participant": "bids_participant_id",
    "session": "bids_session_id",
    "pipeline_name": "pipeline_name",
    "pipeline_version": "pipeline_version",
    "status": "status",
}


def check_if_pipeline_catalog_available():
    """
    Check if the pipeline catalog has been successfully fetched from the remote source.
    """
    if mappings.PIPELINE_CATALOG != []:
        if mappings.PIPELINES_FETCHING_ERR:
            logger.warning(
                f"Failed to fetch pipeline catalog from {mappings.PROCESSING_PIPELINE_URL}. Error: {mappings.PIPELINES_FETCHING_ERR}. "
                "Using a packaged backup pipeline catalog instead *which may be outdated*. "
                "Check your internet connection?"
            )
    else:
        log_error(
            logger,
            f"Failed to load the pipeline catalog. Error: {mappings.PIPELINES_FETCHING_ERR} "
            "Please check that you have an internet connection and try again, or open an issue in https://github.com/neurobagel/bagel-cli/issues if the problem persists.",
        )


def parse_pipeline_catalog(pipeline_catalog: list) -> tuple[dict, dict]:
    """
    Load the pipeline catalog and return a dictionary of pipeline names and their URIs in the Nipoppy namespace,
    and a dictionary of pipeline names and their supported versions in Nipoppy.
    """
    pipeline_versions = {}
    pipeline_uris = {}
    for pipeline in pipeline_catalog:
        pipeline_versions[pipeline["name"]] = pipeline["versions"]
        pipeline_uris[pipeline["name"]] = (
            f"{mappings.NP.pf}:{pipeline['name']}"
        )

    return pipeline_uris, pipeline_versions


def get_recognized_pipelines(
    pipelines: Iterable[str], known_pipeline_uris: dict
) -> list:
    """
    Check that all pipelines in the processing status file are supported by Nipoppy.
    Log an error if all pipelines are unrecognized, otherwise warn about unrecognized pipelines.
    """
    recognized_pipelines = list(
        set(pipelines).intersection(known_pipeline_uris)
    )
    unrecognized_pipelines = list(
        set(pipelines).difference(known_pipeline_uris)
    )

    unrecognized_pipelines_details = (
        f"Unrecognized processing pipelines: {unrecognized_pipelines}. "
        f"Supported pipelines are found in the Nipoppy pipeline catalog (https://github.com/nipoppy/pipeline-catalog): "
        f"{list(known_pipeline_uris.keys())}"
    )
    if not recognized_pipelines:
        log_error(
            logger,
            f"The processing status file contains no recognized pipelines in the column: '{PROC_STATUS_COLS['pipeline_name']}'.\n"
            f"{unrecognized_pipelines_details}",
        )
    if unrecognized_pipelines:
        logger.warning(
            f"The processing status file contains unrecognized pipelines in the column: '{PROC_STATUS_COLS['pipeline_name']}' - "
            "these will be ignored. "
            f"{unrecognized_pipelines_details}"
        )
    return recognized_pipelines


def validate_pipeline_versions(
    pipeline: str, versions: Iterable[str], known_pipeline_versions: dict
) -> tuple[list, list]:
    """
    For a given pipeline, return the recognized and unrecognized pipeline versions in the processing status file
    based on the Nipoppy pipeline catalog, and return both as lists.
    """
    recognized_versions = list(
        set(versions).intersection(known_pipeline_versions[pipeline])
    )
    unrecognized_versions = list(
        set(versions).difference(known_pipeline_versions[pipeline])
    )

    return recognized_versions, unrecognized_versions


def check_at_least_one_pipeline_version_is_recognized(
    status_df: pd.DataFrame,
    known_pipeline_uris: dict,
    known_pipeline_versions: dict,
):
    """
    Check that at least one pipeline name and version combination found in the processing status file is supported by Nipoppy.
    """
    recognized_pipelines = get_recognized_pipelines(
        pipelines=status_df[PROC_STATUS_COLS["pipeline_name"]].unique(),
        known_pipeline_uris=known_pipeline_uris,
    )

    any_recognized_versions = False
    unrecognized_pipeline_versions = {}
    for pipeline in recognized_pipelines:
        versions = status_df[
            status_df[PROC_STATUS_COLS["pipeline_name"]] == pipeline
        ][PROC_STATUS_COLS["pipeline_version"]].unique()

        recognized_versions, unrecognized_versions = (
            validate_pipeline_versions(
                pipeline=pipeline,
                versions=versions,
                known_pipeline_versions=known_pipeline_versions,
            )
        )
        if recognized_versions:
            any_recognized_versions = True
        if unrecognized_versions:
            unrecognized_pipeline_versions[pipeline] = unrecognized_versions

    unrecognized_versions_details = (
        f"Unrecognized processing pipeline versions: {unrecognized_pipeline_versions}. "
        "Supported pipeline versions are found in the Nipoppy pipeline catalog (https://github.com/nipoppy/pipeline-catalog)."
    )
    if not any_recognized_versions:
        log_error(
            logger,
            f"The processing status file contains no recognized versions of any pipelines in the column '{PROC_STATUS_COLS['pipeline_version']}'.\n"
            f"{unrecognized_versions_details}",
        )
    if unrecognized_pipeline_versions:
        logger.warning(
            f"The processing status file contains unrecognized versions of pipelines in the column '{PROC_STATUS_COLS['pipeline_version']}' - "
            "these will be ignored. "
            f"{unrecognized_versions_details}"
        )


def create_completed_pipelines(
    session_proc_df: pd.DataFrame,
    known_pipeline_uris: dict,
    known_pipeline_versions: dict,
) -> list:
    """
    Create a list of CompletedPipeline objects for a single subject-session based on the completion status
    info of pipelines for that session from the processing status dataframe.
    """
    completed_pipelines = []
    for (pipeline, version), session_pipe_df in session_proc_df.groupby(
        [
            PROC_STATUS_COLS["pipeline_name"],
            PROC_STATUS_COLS["pipeline_version"],
        ]
    ):
        if (
            pipeline in known_pipeline_uris
            and version in known_pipeline_versions[pipeline]
        ) and (
            session_pipe_df[PROC_STATUS_COLS["status"]].str.lower()
            == "success"
        ).all():
            completed_pipeline = models.CompletedPipeline(
                hasPipelineName=models.Pipeline(
                    identifier=known_pipeline_uris[pipeline]
                ),
                hasPipelineVersion=version,
            )
            completed_pipelines.append(completed_pipeline)

    return completed_pipelines



from collections import namedtuple
from pathlib import Path

from .utilities import file_utils

DEFAULT_CONFIG = "Neurobagel"

# NOTE: Even though we now support loading of custom namespaces from community configs,
# we cannot remove these hardcoded namespaces yet as they are currently tied to handling logic for specific standardized variables.
# This should be addressed as part of https://github.com/neurobagel/bagel-cli/issues/497.
Namespace = namedtuple("Namespace", ["pf", "url"])
# NOTE: We are keeping the cogatlas namespace for now to warn about its deprecation.
# Remove once we officially support custom vocabularies/namespaces.
COGATLAS = Namespace("cogatlas", "https://www.cognitiveatlas.org/task/id/")
NB = Namespace("nb", "http://neurobagel.org/vocab/")
NCIT = Namespace("ncit", "http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#")
NP = Namespace(
    "np", "https://github.com/nipoppy/pipeline-catalog/tree/main/processing/"
)

# Keep deprecated namespaces for informative user messages
DEPRECATED_NAMESPACES = [COGATLAS]
DEPRECATED_NAMESPACE_PREFIXES = [ns.pf for ns in DEPRECATED_NAMESPACES]

NEUROBAGEL = {
    "participant": NB.pf + ":ParticipantID",
    "session": NB.pf + ":SessionID",
    "sex": NB.pf + ":Sex",
    "age": NB.pf + ":Age",
    "diagnosis": NB.pf + ":Diagnosis",
    "subject_group": NB.pf + ":SubjectGroup",
    "assessment_tool": NB.pf + ":Assessment",
}

# TODO: Use importlib.resources.files(bagel) to get the path to the pipeline-catalog instead?
PROCESSING_PIPELINE_PATH = (
    Path(__file__).parent / "pipeline-catalog/processing/processing.json"
)
PROCESSING_PIPELINE_URL = "https://raw.githubusercontent.com/nipoppy/pipeline-catalog/refs/heads/main/processing/processing.json"
CONFIG_NAMESPACES_PATH = (
    Path(__file__).parent
    / "communities/config_metadata/config_namespace_map.json"
)
CONFIG_NAMESPACES_URL = "https://raw.githubusercontent.com/neurobagel/communities/refs/heads/main/config_metadata/config_namespace_map.json"


# TODO: consider refactoring this into a Mappings class that also
# handles lazy loading of the remote content, i.e. only when accessed
PIPELINE_CATALOG, PIPELINES_FETCHING_ERR = file_utils.request_file(
    url=PROCESSING_PIPELINE_URL, backup_path=PROCESSING_PIPELINE_PATH
)
CONFIG_NAMESPACES_MAPPING, CONFIG_NAMESPACES_FETCHING_ERR = (
    file_utils.request_file(
        url=CONFIG_NAMESPACES_URL, backup_path=CONFIG_NAMESPACES_PATH
    )
)



import inspect
from pathlib import Path

import pydantic

from bagel import mappings, models

EXCLUDE_CLASSES = ["Bagel", "BaseModel"]
PREAMBLE = "\n".join(
    [
        "@prefix nb: <http://neurobagel.org/vocab/>.",
        "@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.",
    ]
)
OUTPUT_PATH = Path.cwd() / "nb_vocab.ttl"


def generate_vocab_string_from_models():
    all_classes = {
        name: obj
        for name, obj in inspect.getmembers(models)
        if inspect.isclass(obj)
        and issubclass(obj, pydantic.BaseModel)
        and name not in EXCLUDE_CLASSES
    }

    class_entries = []
    for name, obj in sorted(all_classes.items()):
        class_entry = f"{mappings.NB.pf}:{name} a rdfs:Class"

        is_subclass = False
        for other_name, other_obj in all_classes.items():
            if other_obj in obj.__bases__:
                class_entry += (
                    f";\n    rdfs:subClassOf {mappings.NB.pf}:{other_name}."
                )
                is_subclass = True
                break

        if not is_subclass:
            class_entry += "."

        class_entries.append(class_entry)

    nb_vocab = "\n\n".join([PREAMBLE] + class_entries) + "\n"

    with OUTPUT_PATH.open("w") as f:
        f.write(nb_vocab)


if __name__ == "__main__":
    generate_vocab_string_from_models()



from functools import lru_cache
from pathlib import Path
from typing import Iterable

import bidsschematools.schema as bst
import pandas as pd
import pandera.pandas as pa
from typer import BadParameter

from bagel import bids_table_model, models
from bagel.logger import log_error, logger
from bagel.utilities import file_utils

# NOTE: A copy of the imaging modality vocab will likely end up in all community config directories,
# but since the contents will be the same, we always pull it from the Neurobagel config for now for simplicity.
IMAGING_MODALITIES_URL = "https://raw.githubusercontent.com/neurobagel/communities/refs/heads/main/configs/Neurobagel/imaging_modalities.json"
IMAGING_MODALITIES_PATH = (
    Path(__file__).parents[1]
    / "communities/configs/Neurobagel/imaging_modalities.json"
)

bids_schema = bst.load_schema()


def get_bids_suffix_to_std_term_mapping() -> dict[str, str]:
    """
    Fetch the standardized imaging modality vocabulary from the neurobagel/communities repository
    and return a mapping of supported BIDS suffixes to prefixed standardized terms.

    Returns:
        dict[str, str]: A mapping where keys are BIDS suffixes (e.g., "T1w", "bold")
                        and values are namespaced standardized terms (e.g., "nidm:T1Weighted").

    Raises:
        typer.Exit: If the vocabulary cannot be fetched from either the remote URL or local backup.
    """
    # TODO: Revisit once we cache community config files locally as part of https://github.com/neurobagel/bagel-cli/issues/493.
    # For now we revert to a local submodule backup if the request fails, but this means we might not have the latest vocab.
    bids_terms_vocab, err = file_utils.request_file(
        url=IMAGING_MODALITIES_URL, backup_path=IMAGING_MODALITIES_PATH
    )
    if bids_terms_vocab == []:
        log_error(
            logger,
            f"Failed to fetch the standardized imaging modality vocabulary required to validate your BIDS metadata. Error: {err} "
            "Please check that you have an internet connection and try again, or open an issue in https://github.com/neurobagel/bagel-cli/issues if the problem persists.",
        )
    # We only expect one term namespace in the bids_term_vocab list
    bids_terms_vocab = bids_terms_vocab[0]
    namespace_prefix = bids_terms_vocab["namespace_prefix"]
    bids_suffix_to_std_term_mapping = {}
    for term in bids_terms_vocab["terms"]:
        bids_suffix_to_std_term_mapping[term["abbreviation"]] = (
            f"{namespace_prefix}:{term['id']}"
        )

    return bids_suffix_to_std_term_mapping


@lru_cache()
def get_all_bids_suffixes() -> set[str]:
    """Return all file suffixes that are recognized by BIDS."""
    return {
        bids_suffix["value"]
        for bids_suffix in bids_schema.objects.suffixes.values()
    }


@lru_cache()
def get_bids_raw_data_suffixes() -> set[str]:
    """Return all BIDS recognized suffixes corresponding to raw data files."""
    bids_raw_data_suffixes = set()
    for file_datatype in bids_schema.rules.files.raw.values():
        for file_subtype in file_datatype.values():
            bids_raw_data_suffixes.update(file_subtype["suffixes"])

    return bids_raw_data_suffixes


def partition_suffixes(
    suffixes: Iterable[str], reference_suffixes: Iterable[str]
) -> tuple[set[str], set[str]]:
    """
    Partition suffixes into those found in a reference collection and those not found.

    Parameters
    ----------
    suffixes : Iterable[str]
        File suffixes to partition.
    reference_suffixes : Iterable[str]
        Suffixes to compare the input list against.

    Returns
    -------
    tuple[set[str], set[str]]
        A tuple containing two sets:
        - The first set contains suffixes found in the reference list.
        - The second set contains suffixes not found in the reference list.
    """
    reference_set = set(reference_suffixes)
    in_reference = set()
    not_in_reference = set()
    for suffix in suffixes:
        if suffix in reference_set:
            in_reference.add(suffix)
        else:
            not_in_reference.add(suffix)
    return in_reference, not_in_reference


def check_absolute_path(dir_path: Path | None) -> Path | None:
    """
    Raise an error if the input path does not look like an absolute path.
    This is a workaround for --dataset-source-path not requiring the path to exist on the host machine
    (and thus not being able to resolve the path automatically).
    """
    # Allow POSIX-style absolute paths (e.g., "/data/...") across OSes - useful for referencing paths on remote Unix servers.
    if dir_path is not None and not (
        dir_path.is_absolute() or dir_path.as_posix().startswith("/")
    ):
        raise BadParameter(
            "Dataset source directory must be an absolute path."
        )
    return dir_path


def validate_bids_table(bids_table: pd.DataFrame):
    """Error and exit if the provided BIDS table is empty or fails schema validation."""
    try:
        bids_table_model.model.validate(bids_table)
    except pa.errors.SchemaError as err:
        rows_with_errs_msg = ""
        # When validation fails due to a column value check (e.g., as opposed to a missing column),
        # printing the row indices helps with debugging, especially for invalid empty values.
        if isinstance(err.failure_cases, pd.DataFrame):
            rows_with_errs_msg = f"Rows with error (0 = first non-header row): {err.failure_cases['index'].tolist()}. "
        log_error(
            logger,
            f"Invalid BIDS table. Error: {err}. {rows_with_errs_msg}",
        )
    if bids_table.empty:
        log_error(
            logger,
            "BIDS table is empty (only a header row was found). No imaging metadata to add.",
        )


def create_acquisitions(
    session_df: pd.DataFrame,
    bids_suffix_term_map: dict,
) -> list:
    """Parses BIDS image file suffixes for a specified session to create a list of Acquisition objects."""
    image_list = []

    for bids_file_suffix in session_df["suffix"]:
        if (
            mapped_term := bids_suffix_term_map.get(bids_file_suffix)
        ) is not None:
            image_list.append(
                models.Acquisition(
                    hasContrastType=models.Image(identifier=mapped_term)
                )
            )

    return image_list


def get_session_path(
    dataset_root: Path | None,
    bids_sub_id: str,
    session_id: str,
) -> str:
    """
    Construct the session directory or subject directory path (when there is no session ID) from the source BIDS directory path, subject ID, and session ID.
    If no source BIDS directory is available, return a relative path.
    """
    subject_path = (
        Path(dataset_root / bids_sub_id) if dataset_root else Path(bids_sub_id)
    )
    session_path = (
        subject_path / session_id if session_id.strip() != "" else subject_path
    )
    return session_path.as_posix()



# This is an example script for generating a JSON schema for the Neurobagel data dictionary model.
# Example usage: python generate_dictionary_jsonschema.py

import json

from bagel.dictionary_models import DataDictionary

FPATH = "neurobagel_datadictionary.schema.json"

with open(FPATH, "w") as f:
    f.write(json.dumps(DataDictionary.model_json_schema(), indent=2))



from enum import Enum
from typing import Annotated, Any

from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    EmailStr,
    Field,
    HttpUrl,
    field_validator,
)


def whitespace_string_to_default_none(value: Any) -> Any:
    """Convert an empty string or a string that contains only whitespace to None."""
    if isinstance(value, str) and value.strip() == "":
        return None
    return value


class AccessType(str, Enum):
    """Enum for dataset access type."""

    PUBLIC = "public"
    REGISTERED = "registered"
    RESTRICTED = "restricted"


class DatasetDescription(BaseModel):
    """Schema for a Neurobagel dataset description JSON file."""

    name: Annotated[
        str,
        Field(
            ...,
            description=(
                "Name of the dataset. This name will be displayed when users discover the dataset in a Neurobagel query. "
                "Key reused from BIDS dataset_description.json."
            ),
            alias="Name",
        ),
    ]
    authors: Annotated[
        list[str],
        Field(
            default_factory=list,
            description="List of individuals who contributed to the creation/curation of the dataset. Key reused from BIDS dataset_description.json.",
            alias="Authors",
        ),
    ]
    # NOTE: In the BIDS dataset_description.json, values for this field can be string references as well as URLs.
    references_and_links: Annotated[
        list[str],
        Field(
            default_factory=list,
            description="List of links or references that contain information on the dataset. Key reused from BIDS dataset_description.json.",
            alias="ReferencesAndLinks",
        ),
    ]
    keywords: Annotated[
        list[str],
        Field(
            default_factory=list,
            description="List of keywords describing the content or subject matter of the dataset. Key reused from BIDS dataset_description.json.",
            alias="Keywords",
        ),
    ]
    repository_url: Annotated[
        HttpUrl | None,
        Field(
            default=None,
            description="URL to a repository where the dataset can be downloaded or retrieved from (e.g., DataLad, Zenodo, GitHub).",
            alias="RepositoryURL",
        ),
        BeforeValidator(whitespace_string_to_default_none),
    ]
    access_instructions: Annotated[
        str | None,
        Field(
            default=None,
            description="Description of how to access the data.",
            alias="AccessInstructions",
        ),
        AfterValidator(whitespace_string_to_default_none),
    ]
    access_type: Annotated[
        AccessType,
        Field(
            default=None,
            description="Level of requirements for dataset access.",
            alias="AccessType",
        ),
    ]
    access_email: Annotated[
        EmailStr | None,
        Field(
            default=None,
            description="Primary email for access requests.",
            alias="AccessEmail",
        ),
        BeforeValidator(whitespace_string_to_default_none),
    ]
    access_link: Annotated[
        HttpUrl | None,
        Field(
            default=None,
            description="Primary link for access requests or information.",
            alias="AccessLink",
        ),
        BeforeValidator(whitespace_string_to_default_none),
    ]

    # NOTE: url_preserve_empty_path (>=2.12) is needed to prevent HttpUrl from auto-appending trailing slashes to URLs
    # when the path portion of the URL is empty
    # (https://docs.pydantic.dev/latest/migration/#url-and-dsn-types-in-pydanticnetworks-no-longer-inherit-from-str)
    model_config = ConfigDict(extra="ignore", url_preserve_empty_path=True)

    @field_validator("name")
    @classmethod
    def check_name_not_whitespace(cls, value: str) -> str:
        """
        Raise an error (will be caught as a ValidationError by Pydantic) if
        the required 'Name' field is an empty string or all whitespace.
        """
        if value.strip() == "":
            raise ValueError("'Name' field cannot be an empty string.")
        return value

    @field_validator("authors", "references_and_links", "keywords")
    @classmethod
    def whitespace_list_to_default_empty_list(
        cls, value: list[str]
    ) -> list[str]:
        """Convert a list containing only empty or whitespace strings to an empty list."""
        if all(item.strip() == "" for item in value):
            return []
        return value



import uuid
from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, EmailStr, Field, HttpUrl

from bagel.dataset_description_model import AccessType
from bagel.mappings import NB

UUID_PATTERN = r"[0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12}$"
BAGEL_UUID_PATTERN = rf"^{NB.pf}:{UUID_PATTERN}"


class Bagel(BaseModel):
    """identifier has to be a valid UUID prepended by the Neurobagel namespace
    by default, a random (uuid4) string UUID will be created"""

    identifier: Annotated[
        str,
        Field(
            pattern=BAGEL_UUID_PATTERN,
            default_factory=lambda: NB.pf + ":" + str(uuid.uuid4()),
        ),
    ]

    # NOTE: url_preserve_empty_path (>=2.12) is needed to prevent HttpUrl from auto-appending trailing slashes to URLs
    # when the path portion of the URL is empty
    # (https://docs.pydantic.dev/latest/migration/#url-and-dsn-types-in-pydanticnetworks-no-longer-inherit-from-str)
    model_config = ConfigDict(extra="forbid", url_preserve_empty_path=True)


class ControlledTerm(BaseModel):
    identifier: str | HttpUrl
    schemaKey: str


class Sex(ControlledTerm):
    schemaKey: Literal["Sex"] = "Sex"


class Diagnosis(ControlledTerm):
    schemaKey: Literal["Diagnosis"] = "Diagnosis"


class SubjectGroup(ControlledTerm):
    schemaKey: Literal["SubjectGroup"] = "SubjectGroup"


class Assessment(ControlledTerm):
    schemaKey: Literal["Assessment"] = "Assessment"


class Image(ControlledTerm):
    schemaKey: Literal["Image"] = "Image"


class Acquisition(Bagel):
    hasContrastType: Image
    schemaKey: Literal["Acquisition"] = "Acquisition"


class Pipeline(ControlledTerm):
    schemaKey: Literal["Pipeline"] = "Pipeline"


class CompletedPipeline(Bagel):
    hasPipelineVersion: str
    hasPipelineName: Pipeline
    schemaKey: Literal["CompletedPipeline"] = "CompletedPipeline"


class Session(Bagel):
    hasLabel: str


class PhenotypicSession(Session):
    hasAge: float | None = None
    hasSex: Sex | None = None
    isSubjectGroup: SubjectGroup | None = None
    hasDiagnosis: list[Diagnosis] | None = None
    hasAssessment: list[Assessment] | None = None
    schemaKey: Literal["PhenotypicSession"] = "PhenotypicSession"


class ImagingSession(Session):
    # NOTE: Do imaging session have to have at least one acquisition OR at least one completed pipeline to be valid?
    hasFilePath: str | None = None
    hasAcquisition: list[Acquisition] | None = None
    hasCompletedPipeline: list[CompletedPipeline] | None = None
    schemaKey: Literal["ImagingSession"] = "ImagingSession"


class Subject(Bagel):
    hasLabel: str
    hasSession: list[PhenotypicSession | ImagingSession]
    schemaKey: Literal["Subject"] = "Subject"


class Dataset(Bagel):
    hasLabel: str
    hasAuthors: list[str] | None = None
    hasReferencesAndLinks: list[str] | None = None
    hasKeywords: list[str] | None = None
    hasRepositoryURL: HttpUrl | None = None
    hasAccessInstructions: str | None = None
    hasAccessType: AccessType | None = None
    hasAccessEmail: EmailStr | None = None
    hasAccessLink: HttpUrl | None = None
    hasSamples: list[Subject]
    schemaKey: Literal["Dataset"] = "Dataset"



import pandas as pd
import pandera.extensions as extensions
import pandera.pandas as pa

NO_WHITESPACE_ERR = (
    "Must be a non-empty value that does not contain only whitespace."
)


@extensions.register_check_method()
def is_not_whitespace(pd_series: pd.Series) -> pd.Series:
    """
    Custom Pandera validation method that checks that column values are not empty strings or whitespace.

    TODO: Should non-empty strings with leading/trailing whitespace also fail validation?
    """
    return pd_series.str.strip() != ""


model = pa.DataFrameSchema(
    # NOTE: Most of the columns will implicitly fail on an empty cell via other checks,
    # but we explicitly check for whitespace-only values to provide a more informative error message.
    {
        "sub": pa.Column(
            str,
            checks=[
                pa.Check.is_not_whitespace(error=NO_WHITESPACE_ERR),
                pa.Check.str_startswith("sub-"),
            ],
            nullable=False,
        ),
        "ses": pa.Column(
            str,
            pa.Check(
                lambda ses: (ses.str.strip() == "")
                | ses.str.startswith("ses-"),
                ignore_na=True,
                error='Session ID must be left empty or start with the "ses-" prefix.',
            ),
            nullable=True,
        ),
        "suffix": pa.Column(
            str,
            pa.Check.is_not_whitespace(error=NO_WHITESPACE_ERR),
            nullable=False,
            # NOTE: We filter out unsupported BIDS suffixes and handle them separately
        ),
        "path": pa.Column(
            str,
            pa.Check.is_not_whitespace(error=NO_WHITESPACE_ERR),
            nullable=False,
        ),
    },
    strict="filter",
)



from pathlib import Path

import bids2table as b2t2
import click
import pandas as pd
import typer
from bids import BIDSLayout, exceptions
from rich.progress import Progress, SpinnerColumn, TextColumn, track

from bagel import mappings, models
from bagel._version import __version__

from .logger import VerbosityLevel, configure_logger, log_error, logger
from .utilities import (
    bids_utils,
    derivative_utils,
    file_utils,
    model_utils,
    pheno_utils,
)
from .utilities.derivative_utils import PROC_STATUS_COLS

OPTION_GROUP_NAMES = {
    "troubleshooting": "Troubleshooting",
    "config": "Configuration",
}

CUSTOM_SESSION_LABEL = "ses-unnamed"

bagel = typer.Typer(
    help=(
        "A command-line tool for creating valid, subject-level instances of the Neurobagel graph data model.\n\n"
        "[red]NOTE: The 'pheno' command must always be run first on a new dataset to generate the input .jsonld file required for the 'bids' command.[/red]"
    ),
    # From https://github.com/tiangolo/typer/issues/201#issuecomment-744151303
    context_settings={"help_option_names": ["--help", "-h"]},
    # NOTE: invoking bagel with no options still results in a non-zero exit code
    # even though the help text is printed and there are no errors,
    # because required options are still detected as missing
    no_args_is_help=True,
    rich_markup_mode="rich",
    epilog=(
        "Run 'bagel COMMAND --help' for more information on a command.\n\n"
        "Or visit the documentation at https://neurobagel.org/user_guide/cli/"
    ),
)


# NOTE: We use a reusable option instead of a callback to avoid the complexity
# of needing to specify the flag before the actual CLI command names
def verbosity_option():
    """Create a reusable verbosity option for commands."""
    return typer.Option(
        VerbosityLevel.INFO,
        "--verbosity",
        "-v",
        callback=configure_logger,
        help="Set the verbosity level of the output. 0 = show errors only; 1 = show errors, warnings, and informational messages; 3 = show all logs, including debug messages.",
        rich_help_panel=OPTION_GROUP_NAMES["troubleshooting"],
    )


def overwrite_option():
    """Create a reusable overwrite option for commands."""
    return typer.Option(
        False,
        "--overwrite",
        "-f",
        help="Overwrite output file if it already exists.",
    )


def show_help(ctx: typer.Context, value: bool):
    """
    Callback to display the command help and exit.
    Adapted from https://github.com/fastapi/typer/discussions/833#discussioncomment-9551792.
    """
    if value:
        typer.echo(ctx.get_help())
        raise typer.Exit()


def help_option():
    """
    Create a reusable help option for commands.
    Used to override the default location of the --help option in the help text
    so that it appears in a different named option group.
    """
    return typer.Option(
        False,
        "--help",
        "-h",
        callback=show_help,
        help="Show this message and exit.",
        rich_help_panel=OPTION_GROUP_NAMES["troubleshooting"],
    )


def show_version(value: bool):
    """Callback to show package version number and exit."""
    if value:
        print(f"bagel {__version__}")
        raise typer.Exit()


@bagel.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        help="Show the version and exit.",
        callback=show_version,
        is_eager=True,
    )
):
    """Handle options for the main CLI command 'bagel'."""
    return


@bagel.command()
def bids2tsv(
    bids_dir: Path = typer.Option(
        ...,
        "--bids-dir",
        "-b",
        help="Path to the BIDS dataset directory.",
        exists=True,
        file_okay=False,
        dir_okay=True,
        resolve_path=True,
    ),
    output: Path = typer.Option(
        "bids.tsv",
        "--output",
        "-o",
        help="Path to save the output .tsv file.",
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    overwrite: bool = overwrite_option(),
    verbosity: VerbosityLevel = verbosity_option(),
    help_: bool = help_option(),
):
    """
    Convert a BIDS dataset into a minimal tabular format (.tsv) containing information about subject, session, suffix (image contrast), and file path for imaging files.
    """
    file_utils.check_overwrite(output, overwrite)

    logger.info(f"Input BIDS directory:  {bids_dir}")

    try:
        # NOTE: If there are no subjects in the BIDS dataset, the validation should fail.
        # The rest of this workflow assumes there's at least one subject in the BIDS dataset.
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            # Add a spinner during BIDS parsing - this spinner disappears once the process is done
            # and won't be present in logs
            progress.add_task(
                "Validating BIDS dataset. This may take a while...", total=None
            )
            BIDSLayout(bids_dir, validate=True)
        logger.info("BIDS validation passed.")
    except exceptions.BIDSValidationError as e:
        log_error(
            logger,
            f"Invalid BIDS dataset at {bids_dir}. Validation error: {e}",
        )

    output_columns = ["sub", "ses", "suffix", "path"]
    dataset_tab = b2t2.index_dataset(bids_dir)
    dataset_df = dataset_tab.to_pandas()

    nb_bids_suffix_term_map = bids_utils.get_bids_suffix_to_std_term_mapping()

    bids_recognized_suffixes, bids_unrecognized_suffixes = (
        bids_utils.partition_suffixes(
            suffixes=dataset_df["suffix"].unique().tolist(),
            reference_suffixes=bids_utils.get_all_bids_suffixes(),
        )
    )
    if bids_unrecognized_suffixes:
        logger.warning(
            f"Files with suffixes not recognized by BIDS were found: {list(bids_unrecognized_suffixes)}. "
            "These will be ignored. "
            "Please refer to the BIDS specification https://bids-specification.readthedocs.io/en/stable/ for file naming conventions."
        )
    # We don't care about and don't warn for valid BIDS file suffixes that don't correspond to raw data files
    bids_raw_data_suffixes, _ = bids_utils.partition_suffixes(
        suffixes=bids_recognized_suffixes,
        reference_suffixes=bids_utils.get_bids_raw_data_suffixes(),
    )
    neurobagel_supported_suffixes, neurobagel_unsupported_suffixes = (
        bids_utils.partition_suffixes(
            suffixes=bids_raw_data_suffixes,
            reference_suffixes=nb_bids_suffix_term_map.keys(),
        )
    )
    if neurobagel_unsupported_suffixes:
        logger.warning(
            f"Data files with valid BIDS suffixes that are not supported by Neurobagel were found: {list(neurobagel_unsupported_suffixes)}. "
            f"These will be ignored. Supported BIDS suffixes: {list(nb_bids_suffix_term_map.keys())}."
        )
    if not neurobagel_supported_suffixes:
        log_error(
            logger,
            f"No image files with supported BIDS suffixes were found in {bids_dir}. "
            "A BIDS metadata table could not be generated. "
            "Please ensure your dataset includes at least one image file with a Neurobagel-supported BIDS suffix "
            f"(Supported suffixes: {list(nb_bids_suffix_term_map.keys())}).",
        )

    dataset_df = dataset_df[
        dataset_df["suffix"].isin(neurobagel_supported_suffixes)
    ].copy()

    dataset_df["path"] = dataset_df.apply(
        lambda row: (Path(row["root"]) / row["path"]).as_posix(), axis=1
    )
    dataset_df = dataset_df[output_columns]

    dataset_df["sub"] = dataset_df["sub"].apply(lambda id: f"sub-{id}")
    dataset_df["ses"] = dataset_df["ses"].apply(
        lambda id: f"ses-{id}" if pd.notna(id) else id
    )

    dataset_df.to_csv(output, sep="\t", index=False)
    logger.info(f"Saved output to:  {output}")


@bagel.command()
def pheno(
    pheno: Path = typer.Option(  # TODO: Rename argument to something clearer, like --tabular.
        ...,
        "--pheno",
        "-t",  # for tabular
        help="Path to a phenotypic .tsv file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    dictionary: Path = typer.Option(
        ...,
        "--dictionary",
        "-d",
        help="Path to the .json data dictionary corresponding to the phenotypic .tsv file.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    dataset_description: Path = typer.Option(
        ...,
        "--dataset-description",
        "-m",  # for metadata
        help="Path to a .json file describing the dataset and access information. "
        "If your dataset is BIDS-compliant, you may reuse the BIDS dataset_description.json here."
        "See the documentation at https://neurobagel.org/user_guide/dataset_description/",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    output: Path = typer.Option(
        "pheno.jsonld",
        "--output",
        "-o",
        help="Path to the output .jsonld file.",
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    config: str = typer.Option(
        mappings.DEFAULT_CONFIG,
        "--config",
        "-c",
        # Solution for providing preset choices taken from https://github.com/fastapi/typer/issues/182#issuecomment-1708245110
        # NOTE: Alternatively, we could dynamically create a string listing the available config names
        # to include in the option description in the help text. We would then use a callback to validate the config name manually,
        # instead of using click.Choice which handles displaying the choices and validation automatically.
        # This might be useful once/if we have many community configurations to choose from or want more flexibility in errors.
        click_type=click.Choice(
            pheno_utils.get_available_configs(
                mappings.CONFIG_NAMESPACES_MAPPING
            ),
            case_sensitive=False,
        ),
        help="Name of the vocabulary configuration used to generate your data dictionary in the annotation tool. "
        "If you are processing data for a Neurobagel subcommunity, choose the subcommunity name here. "
        f"{pheno_utils.additional_config_help_text()}",
        rich_help_panel=OPTION_GROUP_NAMES["config"],
    ),
    overwrite: bool = overwrite_option(),
    verbosity: VerbosityLevel = verbosity_option(),
    help_: bool = help_option(),
):
    """
    Process a tabular phenotypic file (.tsv) that has been successfully annotated with the Neurobagel annotation tool, with annotations stored in a data dictionary (.json).

    This command will create a valid, subject-level instance of the Neurobagel graph data model for the provided phenotypic file in the JSON-LD format. You can upload this .jsonld file to the Neurobagel graph.
    """
    file_utils.check_overwrite(output, overwrite)

    data_dictionary = file_utils.load_json(dictionary)
    pheno_df = file_utils.load_tabular(pheno)
    dataset_metadata = file_utils.load_json(dataset_description)

    pheno_utils.check_if_remote_config_namespaces_used()

    logger.info("Running initial checks of inputs...")
    # NOTE: `width` determines the amount of padding (in num. characters) before the file paths in the print statement.
    # It is arbitrarily calculated as = length of the longest string + 2 extra spaces
    width = 30
    logger.info("%-*s%s", width, "Tabular file (.tsv):", pheno)
    logger.info("%-*s%s", width, "Data dictionary (.json):", dictionary)
    logger.info(
        "%-*s%s", width, "Dataset description (.json):", dataset_description
    )
    pheno_utils.validate_inputs(data_dictionary, pheno_df, config)
    dataset_metadata = pheno_utils.validate_dataset_description(
        dataset_metadata
    )

    # TODO: Remove once we no longer support annotation tool v1 data dictionaries
    data_dictionary = pheno_utils.convert_transformation_to_format(
        data_dictionary
    )

    logger.info("Processing phenotypic annotations...")
    subject_list = []

    column_mapping = pheno_utils.map_std_vars_to_columns(data_dictionary)
    tool_mapping = pheno_utils.map_tools_to_columns(data_dictionary)

    # TODO: needs refactoring once we handle multiple participant IDs
    participants = column_mapping["participant"][0]

    # Note that `session_column will be None if there is no session column in the pheno.tsv
    session_column = column_mapping.get("session")

    for participant in pheno_df[participants].unique():
        _sub_pheno = pheno_df.query(
            f"`{participants}` == '{str(participant)}'"
        )

        sessions = []
        for session_row_idx, session_row in _sub_pheno.iterrows():
            # Our data model requires a session. To support phenotypic data without sessions,
            # we create a session with a fixed, but unusual CUSTOM_SESSION_LABEL and add the
            # phenotypic data to that session.
            if session_column is None:
                session_label = CUSTOM_SESSION_LABEL
            else:
                # NOTE: We take the name from the first session column - we don't know how to handle multiple session columns yet
                session_label = session_row[session_column[0]]

            session = models.PhenotypicSession(hasLabel=str(session_label))
            _ses_pheno = session_row

            if "sex" in column_mapping.keys():
                _sex_vals = pheno_utils.get_transformed_values(
                    column_mapping["sex"], _ses_pheno, data_dictionary
                )
                if _sex_vals:
                    # NOTE: Our data model only allows a single sex value, so we only take the first instance if multiple columns are about sex
                    session.hasSex = models.Sex(identifier=_sex_vals[0])

            if "diagnosis" in column_mapping.keys():
                _dx_vals = pheno_utils.get_transformed_values(
                    column_mapping["diagnosis"], _ses_pheno, data_dictionary
                )
                if _dx_vals:
                    session.hasDiagnosis = [
                        models.Diagnosis(identifier=_dx_val)
                        for _dx_val in _dx_vals
                    ]

            if "subject_group" in column_mapping.keys():
                _group_vals = pheno_utils.get_transformed_values(
                    column_mapping["subject_group"],
                    _ses_pheno,
                    data_dictionary,
                )
                if _group_vals:
                    session.isSubjectGroup = models.SubjectGroup(
                        identifier=_group_vals[0]
                    )

            if "age" in column_mapping.keys():
                # NOTE: At the moment, our data model only supports a single age value per subject.
                # To achieve this, we transform the values from ALL columns annotated as about age
                # (so we expect each of them to be valid according to the data dictionary model),
                # but we take and store only the first instance in the graph data.
                _age_vals = pheno_utils.get_transformed_values(
                    column_mapping["age"], _ses_pheno, data_dictionary
                )
                if _age_vals:
                    session.hasAge = _age_vals[0]

            if tool_mapping:
                _assessments = [
                    models.Assessment(identifier=tool)
                    for tool, columns in tool_mapping.items()
                    if pheno_utils.are_any_available(
                        columns, _ses_pheno, data_dictionary
                    )
                ]
                if _assessments:
                    # Only set assessments for the subject if at least one has a non-missing item
                    session.hasAssessment = _assessments
            sessions.append(session)

        subject = models.Subject(
            hasLabel=str(participant), hasSession=sessions
        )
        subject_list.append(subject)

    dataset_graph_attributes = (
        pheno_utils.dataset_description_to_graph_attributes(dataset_metadata)
    )

    dataset = models.Dataset(
        **dataset_graph_attributes,
        hasSamples=subject_list,
    )

    file_utils.save_jsonld(
        data=model_utils.dataset_to_jsonld(
            context=model_utils.generate_context(config),
            dataset=dataset,
        ),
        filename=output,
    )


@bagel.command()
def bids(
    # TODO: Rename to --jsonld? Since other file options do not have the _path suffix.
    jsonld_path: Path = typer.Option(
        ...,
        "--jsonld-path",
        "-p",  # for pheno
        help="Path to the .jsonld file containing the phenotypic data for your dataset, created using the 'bagel pheno' command. "
        "This file may optionally also include the processing pipeline metadata for the dataset (created using the 'bagel derivatives' command).",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    bids_table: Path = typer.Option(
        ...,
        "--bids-table",
        "-b",
        help="Path to a .tsv file containing the BIDS metadata for image files including 'sub', 'ses', 'suffix', and 'path' columns. "
        "This file can be created using the 'bagel bids2tsv' command.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    # TODO: Should we include a tip in the help text for using the repository root for DataLad datasets?
    # NOTE: dataset_source_path or dataset_root_path?
    dataset_source_dir: Path = typer.Option(
        None,
        "--dataset-source-dir",
        "-s",
        callback=bids_utils.check_absolute_path,
        help="Absolute path to the root directory of the BIDS dataset at the source location/file server. "
        "If provided, this path will be combined with the subject and session IDs from the BIDS table "
        "to create absolute source paths to the imaging data for each subject and session.",
        exists=False,
        file_okay=False,
        dir_okay=True,
        resolve_path=False,
    ),
    # TODO: Should we rename the default output file to something more generic to account for the fact that
    # the file may also include derivatives data? e.g., dataset_bids.jsonld
    output: Path = typer.Option(
        "bids.jsonld",
        "--output",
        "-o",
        help="Path to the output .jsonld file.",
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    overwrite: bool = overwrite_option(),
    verbosity: VerbosityLevel = verbosity_option(),
    help_: bool = help_option(),
):
    """
    Extract and integrate imaging metadata from a BIDS dataset with harmonized subject phenotypic data (from 'bagel pheno') and, optionally, processing pipeline metadata (from 'bagel derivatives') in a single .jsonld file.
    [red]NOTE: Must be run AFTER 'bagel pheno'.[/red]

    This command will create a valid, subject-level instance of the Neurobagel graph data model for the combined metadata in the JSON-LD format. You can upload this .jsonld file to the Neurobagel graph.
    """

    file_utils.check_overwrite(output, overwrite)

    width = 51
    logger.info("Running initial checks of inputs...")
    logger.info(
        "%-*s%s",
        width,
        "Existing subject graph data to augment (.jsonld):",
        jsonld_path,
    )
    logger.info("%-*s%s", width, "BIDS dataset table (.tsv):", bids_table)
    if dataset_source_dir:
        logger.info(
            "%-*s%s",
            width,
            "Dataset root directory at the data source location (used to construct imaging data paths):",
            dataset_source_dir,
        )

    jsonld_context, jsonld_dataset = (
        model_utils.extract_and_validate_jsonld_dataset(jsonld_path)
    )
    bids_dataset = file_utils.load_tabular(bids_table, input_type="BIDS")

    nb_bids_suffix_term_map = bids_utils.get_bids_suffix_to_std_term_mapping()
    # NOTE: The BIDS table model validation will check for required columns and for empty values in the "suffix" column
    # and error out for any problem. Because we want to ignore unsupported suffixes with a warning instead of a validation error,
    # we check the suffix column separately here and then remove any offending values.
    # For our custom suffix-check to work, we need to ensure that the "suffix" column exists here.
    if "suffix" in bids_dataset.columns:
        # We assume that most input BIDS TSVs will have been generated by the 'bids2tsv' command,
        # so we only log a generic warning here for any suffixes found in the table that are not supported by Neurobagel
        neurobagel_supported_suffixes, neurobagel_unsupported_suffixes = (
            bids_utils.partition_suffixes(
                suffixes=bids_dataset["suffix"],
                reference_suffixes=nb_bids_suffix_term_map.keys(),
            )
        )
        if not neurobagel_supported_suffixes:
            log_error(
                logger,
                f"No Neurobagel-supported BIDS suffixes found in BIDS table 'suffix' column: {bids_table}. "
                "No imaging metadata could be added to the subject graph data. "
                "Please ensure your dataset includes at least one image file with a Neurobagel-supported BIDS suffix "
                f"(supported suffixes: {list(nb_bids_suffix_term_map.keys())}).",
            )
        if neurobagel_unsupported_suffixes:
            logger.warning(
                f"BIDS table 'suffix' column contains file suffixes unsupported by Neurobagel: {list(neurobagel_unsupported_suffixes)}. "
                f"These records will be ignored. Supported file suffixes: {list(nb_bids_suffix_term_map.keys())}."
            )

        bids_dataset = bids_dataset[
            bids_dataset["suffix"].isin(neurobagel_supported_suffixes)
        ].copy()

    bids_utils.validate_bids_table(bids_dataset)

    existing_subs_dict = model_utils.get_subject_instances(jsonld_dataset)
    bids_subject_ids = list(bids_dataset["sub"].unique())

    model_utils.confirm_subs_match_pheno_data(
        subjects=bids_subject_ids,
        subject_source_for_err="BIDS dataset file",
        pheno_subjects=existing_subs_dict.keys(),
    )

    logger.info("Initial checks of inputs passed.")

    logger.info(
        "Subject metadata for the following Neurobagel-supported imaging modalities "
        f"will be added to the subject graph data: {list(neurobagel_supported_suffixes)}"
    )

    logger.info("Merging BIDS metadata with existing subject annotations...")

    disable_progress_bar = verbosity == VerbosityLevel.ERROR
    for bids_sub_id in track(
        bids_subject_ids,
        description="Processing BIDS subjects...",
        disable=disable_progress_bar,
    ):
        _bids_sub = bids_dataset[bids_dataset["sub"] == bids_sub_id]
        existing_subject = existing_subs_dict[bids_sub_id]
        existing_sessions_dict = model_utils.get_imaging_session_instances(
            existing_subject
        )

        bids_sessions = list(_bids_sub["ses"].unique())
        # TODO: Do we need to explicitly preprocess cases where ses values are other types of whitespace?
        # Ensure the sessions are in alphanumeric order for readability
        for session_id in sorted(bids_sessions):
            _bids_session = _bids_sub[_bids_sub["ses"] == session_id]
            image_list = bids_utils.create_acquisitions(
                session_df=_bids_session,
                bids_suffix_term_map=nb_bids_suffix_term_map,
            )

            if not image_list:
                continue

            # TODO: Currently if a subject has BIDS data but no "ses-" directories (e.g., only 1 session),
            # we create a session with a fixed, but unusual CUSTOM_SESSION_LABEL
            # and add the imaging data info to a session with that label (or create it first).
            # However, we still provide the BIDS SUBJECT directory as the session path, instead of making up a path.
            # This should be revisited in the future as for these cases the resulting dataset object is not
            # an exact representation of what's on disk.
            session_label = (
                CUSTOM_SESSION_LABEL
                if session_id.strip() == ""
                else session_id
            )
            session_path = bids_utils.get_session_path(
                dataset_root=dataset_source_dir,
                bids_sub_id=bids_sub_id,
                session_id=session_id,
            )

            # If a custom Neurobagel-created session already exists (if `bagel derivatives` was run first),
            # we add to that session when there is no session layer in the BIDS directory
            if session_label in existing_sessions_dict:
                existing_img_session = existing_sessions_dict[session_label]
                existing_img_session.hasAcquisition = image_list
                existing_img_session.hasFilePath = session_path
            else:
                new_imaging_session = models.ImagingSession(
                    hasLabel=session_label,
                    hasFilePath=session_path,
                    hasAcquisition=image_list,
                )
                existing_subject.hasSession.append(new_imaging_session)

    # NOTE: We currently reuse the context from the input JSONLD instead of regenerating it to avoid
    # asking the user to specify a config for each command.
    # However, this means that we are not fully protected against the (hopefully rare) case where the context has changed between
    # the generation of the input JSONLD and when this command is run (e.g., if the data model underwent an update in the interim).
    # This may be resolved with https://github.com/neurobagel/bagel-cli/issues/492.
    file_utils.save_jsonld(
        data=model_utils.dataset_to_jsonld(
            context={"@context": jsonld_context},
            dataset=jsonld_dataset,
        ),
        filename=output,
    )


@bagel.command()
def derivatives(
    tabular: Path = typer.Option(
        ...,
        "--tabular",
        "-t",
        help="Path to a .tsv containing subject-level processing pipeline status info. Expected to comply with the Nipoppy processing status file schema.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    # TODO: Remove _path?
    jsonld_path: Path = typer.Option(
        ...,
        "--jsonld-path",
        "-p",  # for pheno
        help="Path to a .jsonld file containing the phenotypic data for your dataset, created using the 'bagel pheno' command. This file may optionally also include the BIDS metadata for the dataset (created using the 'bagel bids' command).",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    output: Path = typer.Option(
        "derivatives.jsonld",
        "--output",
        "-o",
        help="Path to the output .jsonld file.",
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    overwrite: bool = overwrite_option(),
    verbosity: VerbosityLevel = verbosity_option(),
    help_: bool = help_option(),
):
    """
    Extract subject processing pipeline and derivative metadata from a tabular processing status file and integrate them in a single .jsonld with harmonized subject phenotypic data (from 'bagel pheno') and optionally, BIDS metadata (from 'bagel bids').
    [red]NOTE: Must be run AFTER 'bagel pheno'.[/red]

    This command will create a valid, subject-level instance of the Neurobagel graph data model for the combined metadata in the JSON-LD format. You can upload this .jsonld file to the Neurobagel graph.
    """

    file_utils.check_overwrite(output, overwrite)
    derivative_utils.check_if_pipeline_catalog_available()

    width = 51
    logger.info("Processing subject-level derivative metadata...")
    logger.info(
        "%-*s%s",
        width,
        "Existing subject graph data to augment (.jsonld):",
        jsonld_path,
    )
    logger.info("%-*s%s", width, "Processing status file (.tsv):", tabular)

    status_df = file_utils.load_tabular(
        tabular, input_type="processing status"
    )

    # We don't allow empty values in the participant ID column
    if row_indices := pheno_utils.get_rows_with_empty_strings(
        status_df, [PROC_STATUS_COLS["participant"]]
    ):
        log_error(
            logger,
            f"Your processing status file contains missing values in the column '{PROC_STATUS_COLS['participant']}'. "
            "Please ensure that every row has a non-empty participant id. "
            f"We found missing values in the following rows (first row is zero): {row_indices}.",
        )

    known_pipeline_uris, known_pipeline_versions = (
        derivative_utils.parse_pipeline_catalog(mappings.PIPELINE_CATALOG)
    )

    derivative_utils.check_at_least_one_pipeline_version_is_recognized(
        status_df=status_df,
        known_pipeline_uris=known_pipeline_uris,
        known_pipeline_versions=known_pipeline_versions,
    )

    jsonld_context, jsonld_dataset = (
        model_utils.extract_and_validate_jsonld_dataset(jsonld_path)
    )

    existing_subs_dict = model_utils.get_subject_instances(jsonld_dataset)

    model_utils.confirm_subs_match_pheno_data(
        subjects=status_df[PROC_STATUS_COLS["participant"]].unique(),
        subject_source_for_err="processing status file",
        pheno_subjects=existing_subs_dict.keys(),
    )

    # Create sub-dataframes for each subject
    for subject, sub_proc_df in status_df.groupby(
        PROC_STATUS_COLS["participant"]
    ):
        existing_subject = existing_subs_dict[subject]

        # Note: Dictionary of existing imaging sessions can be empty if only bagel pheno was run
        existing_sessions_dict = model_utils.get_imaging_session_instances(
            existing_subject
        )

        for session_label, sub_ses_proc_df in sub_proc_df.groupby(
            PROC_STATUS_COLS["session"]
        ):
            completed_pipelines = derivative_utils.create_completed_pipelines(
                session_proc_df=sub_ses_proc_df,
                known_pipeline_uris=known_pipeline_uris,
                known_pipeline_versions=known_pipeline_versions,
            )

            if not completed_pipelines:
                continue

            session_label = (
                CUSTOM_SESSION_LABEL if session_label == "" else session_label
            )
            if session_label in existing_sessions_dict:
                existing_img_session = existing_sessions_dict[session_label]
                existing_img_session.hasCompletedPipeline = completed_pipelines
            else:
                new_img_session = models.ImagingSession(
                    hasLabel=session_label,
                    hasCompletedPipeline=completed_pipelines,
                )
                existing_subject.hasSession.append(new_img_session)

    # NOTE: We currently reuse the context from the input JSONLD instead of regenerating it to avoid
    # asking the user to specify a config for each command.
    # However, this means that we are not fully protected against the (hopefully rare) case where the context has changed between
    # the generation of the input JSONLD and when this command is run (e.g., if the data model underwent an update in the interim).
    # This may be resolved with https://github.com/neurobagel/bagel-cli/issues/492.
    file_utils.save_jsonld(
        data=model_utils.dataset_to_jsonld(
            context={"@context": jsonld_context},
            dataset=jsonld_dataset,
        ),
        filename=output,
    )


@bagel.command()
def harmonize_pheno(
    pheno: Path = typer.Option(  # TODO: Rename argument to something clearer, like --tabular.
        ...,
        "--pheno",
        "-t",  # for tabular
        help="Path to a phenotypic .tsv file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    dictionary: Path = typer.Option(
        ...,
        "--dictionary",
        "-d",
        help="Path to the .json data dictionary corresponding to the phenotypic .tsv file.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    output: Path = typer.Option(
        "harmonized.tsv",
        "--output",
        "-o",
        help="Path to save the output harmonized .tsv file.",
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    overwrite: bool = overwrite_option(),
    verbosity: VerbosityLevel = verbosity_option(),
    help_: bool = help_option(),
):
    """
    [dim][red]Experimental command, please avoid using in production unless explicitly advised![/red]
    Harmonize the contents of a tabular phenotypic file (.tsv) using the annotations defined in its corresponding Neurobagel data dictionary (.json, generated using the Neurobagel annotation tool).[/dim]
    """
    file_utils.check_overwrite(output, overwrite)

    data_dictionary = file_utils.load_json(dictionary)
    pheno_df = file_utils.load_tabular(pheno)

    logger.info("Running initial checks of inputs...")
    # NOTE: `width` determines the amount of padding (in num. characters) before the file paths in the print statement.
    # It is calculated as = length of the longer string + 2 extra spaces
    width = 26
    logger.info("%-*s%s", width, "Tabular file (.tsv):", pheno)
    logger.info("%-*s%s", width, "Data dictionary (.json):", dictionary)
    pheno_utils.validate_inputs(data_dictionary, pheno_df)

    # TODO: Remove once we no longer support annotation tool v1 data dictionaries
    data_dictionary = pheno_utils.convert_transformation_to_format(
        data_dictionary
    )

    logger.info("Harmonizing raw tabular file...")

    column_mapping = pheno_utils.map_std_vars_to_columns(data_dictionary)
    collection_mapping = pheno_utils.map_tools_to_columns(data_dictionary)

    # Prepare list of annotated columns that will be harmonized and included in the output table
    cols_to_harmonize = []
    for std_var_name, columns in column_mapping.items():
        if std_var_name == "assessment_tool":
            cols_to_harmonize.extend(columns)
        else:
            # NOTE: By default, we use only the first column mapped to a specific std_var
            # to avoid duplicate harmonized column names in the output TSV.
            cols_to_harmonize.append(columns[0])

    transformed_rows = []
    for _, row in pheno_df.iterrows():
        transformed_row = pheno_utils.get_transformed_row_for_table(
            cols_to_harmonize, row, data_dictionary, collection_mapping
        )
        transformed_rows.append(transformed_row)
    harmonized_pheno_df = pd.DataFrame(transformed_rows)

    harmonized_pheno_df.to_csv(output, sep="\t", index=False)
    logger.info(f"Saved harmonized table to:  {output}")



import inspect
from pathlib import Path
from typing import Iterable

import pydantic
from pydantic import ValidationError

from bagel import models
from bagel.logger import log_error, logger
from bagel.mappings import NB
from bagel.utilities import file_utils, pheno_utils


def generate_context(config: str) -> dict:
    # Adapted from the dandi-schema context generation function
    # https://github.com/dandi/dandi-schema/blob/c616d87eaae8869770df0cb5405c24afdb9db096/dandischema/metadata.py
    field_preamble = pheno_utils.get_supported_namespaces_for_config(config)
    fields: dict[str, str | dict[str, str]] = {}
    for klass_name, klass in inspect.getmembers(models):
        if inspect.isclass(klass) and issubclass(klass, pydantic.BaseModel):
            fields[klass_name] = f"{NB.pf}:{klass_name}"
            for name, field in klass.model_fields.items():
                if name == "schemaKey":
                    fields[name] = "@type"
                elif name == "identifier":
                    fields[name] = "@id"
                elif name not in fields:
                    fields[name] = {"@id": f"{NB.pf}:{name}"}

    field_preamble.update(**fields)

    return {"@context": field_preamble}


def dataset_to_jsonld(context: dict, dataset: models.Dataset) -> dict:
    """Add the Neurobagel context to a graph-ready dataset to form a JSONLD dictionary."""
    # We can't just exclude_unset here because the identifier and schemaKey
    # for each instance are created as default values and so technically are never set
    # TODO: we should revisit this because there may be reasons to have None be meaningful in the future
    return {**context, **dataset.model_dump(exclude_none=True, mode="json")}


def get_subs_missing_from_pheno_data(
    subjects: Iterable, pheno_subjects: Iterable
) -> list:
    """Check a list of subject IDs and return any not found in the provided phenotypic subject list."""
    return list(set(subjects).difference(pheno_subjects))


def confirm_subs_match_pheno_data(
    subjects: Iterable, subject_source_for_err: str, pheno_subjects: Iterable
):
    """
    Return an error if not all subjects in the subject list are found in the provided phenotypic subject list.
    """
    missing_subs = get_subs_missing_from_pheno_data(
        subjects=subjects,
        pheno_subjects=pheno_subjects,
    )

    if len(missing_subs) > 0:
        log_error(
            logger,
            f"The {subject_source_for_err} contains subject IDs not found in "
            "the provided JSON-LD file:\n"
            f"{missing_subs}\n"
            "Subject IDs are case sensitive. "
            f"Please check that the {subject_source_for_err} corresponds to the dataset in the .jsonld file.",
        )


def extract_and_validate_jsonld_dataset(
    file_path: Path,
) -> tuple[dict, models.Dataset]:
    """
    Separate the context out from a user-provided JSONLD and validate the remaining contents
    against the data model for a Neurobagel dataset.
    """
    jsonld = file_utils.load_json(file_path)
    context = jsonld.pop("@context")
    try:
        jsonld_dataset = models.Dataset.model_validate(jsonld)
    except ValidationError as err:
        log_error(
            logger,
            f"Error: {file_path} is not a valid Neurobagel JSONLD dataset. "
            "Please ensure to provide a valid JSONLD file generated by Neurobagel CLI commands.\n"
            f"Validation errors: {str(err)}",
        )

    return context, jsonld_dataset


def get_subject_instances(
    dataset: models.Dataset,
) -> dict[str, models.Subject]:
    """
    Return a dictionary of subjects for a given Neurobagel dataset from JSONLD data,
    where keys are subject labels and values are the subject objects.
    """
    return {
        subject.hasLabel: subject for subject in getattr(dataset, "hasSamples")
    }


def get_imaging_session_instances(
    jsonld_subject: models.Subject,
) -> dict[str, models.ImagingSession]:
    """
    Return a dictionary of imaging sessions for a given subject from JSONLD data,
    where the keys are the session labels and values are the session objects.
    """
    jsonld_sub_sessions_dict = {}
    for jsonld_sub_ses in getattr(jsonld_subject, "hasSession"):
        if jsonld_sub_ses.schemaKey == "ImagingSession":
            jsonld_sub_sessions_dict[jsonld_sub_ses.hasLabel] = jsonld_sub_ses

    return jsonld_sub_sessions_dict



import logging
import sys
from enum import Enum
from typing import NoReturn

import typer
from rich.logging import RichHandler

LOG_FMT = "%(message)s"
DATETIME_FMT = "[%Y-%m-%d %X]"

# Check if code is currently running in a test environment
IS_TESTING = "pytest" in sys.modules

logger = logging.getLogger("bagel.logger")


class VerbosityLevel(str, Enum):
    """Enum for verbosity levels."""

    ERROR = "0"
    INFO = "1"
    # NOTE: We currently do not have any debug logs, but can add them in the future
    DEBUG = "2"


verbosity_log_levels = {
    VerbosityLevel.ERROR: logging.ERROR,
    VerbosityLevel.INFO: logging.INFO,
    VerbosityLevel.DEBUG: logging.DEBUG,
}


def configure_logger(
    verbosity: VerbosityLevel = VerbosityLevel.INFO,
) -> str:
    """Configure a logger with the specified logging level."""
    level = verbosity_log_levels[verbosity]

    # Prevent duplicate handlers when updating the logger
    if not logger.handlers:
        handler = RichHandler(
            omit_repeated_times=False,
            show_path=False,
            rich_tracebacks=True,
            markup=True,
        )
        formatter = logging.Formatter(fmt=LOG_FMT, datefmt=DATETIME_FMT)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)

    # propagate should be False to avoid duplicate messages from root logger
    # except when testing because otherwise Pytest does not capture the logs
    if not IS_TESTING:
        logger.propagate = False

    # NOTE: We must return the value instead of the level itself to avoid the callback returning None,
    # see bug tracked in https://github.com/fastapi/typer/issues/223
    return verbosity.value


def log_error(
    logger: logging.Logger,
    message: str,
) -> NoReturn:
    """Log an exception with an informative error message, and exit the app."""
    logger.error(message)
    raise typer.Exit(code=1)



from typing import Annotated, Literal

from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
)
from pydantic_core import PydanticCustomError


def validate_unique_list(values: list[str]) -> list[str]:
    """
    Check that provided list only has unique elements.

    This custom validator is needed because constrained dtypes and their `unique_items` parameter
    were deprecated in Pydantic v2. This function was adapted from https://github.com/pydantic/pydantic-core/pull/820#issuecomment-1656228704
    and https://docs.pydantic.dev/latest/concepts/validators/#annotated-validators.

    See also:
    - https://docs.pydantic.dev/latest/migration/#changes-to-pydanticfield
    - https://docs.pydantic.dev/latest/api/types/#pydantic.types.conlist)
    """
    if len(values) != len(set(values)):
        raise PydanticCustomError(
            "unique_list", f"{values} is not a unique list"
        )
    return values


class Term(BaseModel):
    """An identifier of a controlled term with an IRI"""

    termURL: Annotated[
        str,
        Field(
            ...,
            description="An unambiguous identifier for the term, concept or entity that is referenced",
            alias="TermURL",
        ),
    ]
    label: Annotated[
        str,
        Field(
            ...,
            description="A human readable label. If more than one label exists for the term, "
            "then the preferred label should be used.",
            alias="Label",
        ),
    ]


class Neurobagel(BaseModel):
    """The base model for a Neurobagel column annotation"""

    isAbout: Annotated[
        Term,
        Field(
            ...,
            description="The concept or controlled term that describes this column",
            alias="IsAbout",
        ),
    ]
    missingValues: Annotated[
        list[str],
        AfterValidator(validate_unique_list),
        Field(
            [],
            description="A list of unique values that represent "
            "invalid responses, typos, or missing data",
            alias="MissingValues",
            json_schema_extra={"uniqueItems": True},
        ),
    ]

    model_config = ConfigDict(extra="forbid")


class IdentifierNeurobagel(BaseModel):
    """A Neurobagel annotation for an identifier column"""

    isAbout: Annotated[
        Term,
        Field(
            ...,
            description="The concept or controlled term that describes this column",
            alias="IsAbout",
        ),
    ]
    variableType: Annotated[
        Literal["Identifier"], Field(..., alias="VariableType")
    ]

    model_config = ConfigDict(extra="forbid")


class CategoricalNeurobagel(Neurobagel):
    """A Neurobagel annotation for a categorical column"""

    levels: Annotated[
        dict[str, Term],
        Field(
            ...,
            description="For categorical variables: "
            "An object of values (keys) in the column and the semantic "
            "term (URI and label) they are unambiguously mapped to.",
            alias="Levels",
        ),
    ]
    variableType: Annotated[
        Literal["Categorical"], Field(..., alias="VariableType")
    ]


class ContinuousNeurobagel(Neurobagel):
    """A Neurobagel annotation for a continuous column"""

    format: Annotated[
        Term,
        Field(
            ...,
            description="For continuous columns this field is used to describe "
            "the format of the raw numerical values in the column. This information is used to transform "
            "the column values into the desired format of the standardized "
            "data element referenced in the IsAbout attribute.",
            alias="Format",
        ),
    ]
    variableType: Annotated[
        Literal["Continuous"], Field(..., alias="VariableType")
    ]


class CollectionNeurobagel(Neurobagel):
    """
    A Neurobagel annotation for a column that is part of a grouped collection of columns,
    such as items from an instrument.
    """

    isPartOf: Annotated[
        Term,
        Field(
            ...,
            description="If the column is a subscale or item of an assessment tool "
            "then the assessment tool should be specified here.",
            alias="IsPartOf",
        ),
    ]
    variableType: Annotated[
        Literal["Collection"], Field(..., alias="VariableType")
    ]


class Column(BaseModel):
    """The base model for a BIDS column description"""

    # TODO: Revisit if we want to make description an optional field, since we don't currently use it in the graph data.
    # At the moment, the key itself is always required and the value can be an empty string "",
    # but a value of null ("Description": null) is invalid and will result in a schema validation error.
    description: Annotated[
        str,
        Field(
            ...,
            description="Free-form natural language description",
            alias="Description",
        ),
    ]
    annotations: Annotated[
        CategoricalNeurobagel
        | ContinuousNeurobagel
        | IdentifierNeurobagel
        | CollectionNeurobagel,
        Field(None, description="Semantic annotations", alias="Annotations"),
    ]


class CategoricalColumn(Column):
    """A BIDS column annotation for a categorical column"""

    levels: Annotated[
        dict[str, str],
        Field(
            ...,
            description="For categorical variables: "
            "An object of possible values (keys) "
            "and their descriptions (values). ",
            alias="Levels",
        ),
    ]


class ContinuousColumn(Column):
    """A BIDS column annotation for a continuous column"""

    units: Annotated[
        str,
        Field(
            ...,
            description="Measurement units for the values in this column. "
            "SI units in CMIXF formatting are RECOMMENDED (see Units)",
            alias="Units",
        ),
    ]


class DataDictionary(
    RootModel[dict[str, Column | ContinuousColumn | CategoricalColumn]]
):
    """A data dictionary with human and machine readable information for a tabular data file"""

    pass
